# -*- coding: utf-8 -*-
"""backup module"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import json
import logging
import os
import requests
import time
import yaml

from datetime import datetime
from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.modules import DWhoModuleBase, MODULES
from dwho.config import DWHO_SHARED
from mako.template import Template
from sonicprobe.libs import urisup, xys
from sonicprobe.libs.http_json_server import HttpReqErrJson
from sonicprobe.libs.moresynchro import RWLock


requests.packages.urllib3.disable_warnings()


LOG        = logging.getLogger('xtrargon.modules.backup')


class XtrargonBackupModule(DWhoModuleBase):
    MODULE_NAME = 'backup'

    LOCK        = RWLock()

    def safe_init(self, options):
        self.server_id     = self.config['general']['server_id']
        self.lock_timeout  = self.config['general']['lock_timeout']
        self.adapter_redis = DWhoAdapterRedis(self.config, prefix = 'supervision')
        self.notifiers     = {}
        DWHO_SHARED.add('xtrargon')

        self._get_notifiers()

        if not self.adapter_redis.servers:
            raise SystemError("missing redis server")

    def _get_notifiers(self):
        for xfile in os.listdir(self.config['general']['notifiers_path']):
            xpath = os.path.join(self.config['general']['notifiers_path'], xfile)
            if not xpath.endswith('.yml') or not os.path.isfile(xpath):
                continue

            f = None
            with open(xpath, 'r') as f:
                name = os.path.splitext(os.path.basename(xpath))[0]
                cfg  = yaml.load(f)

                self.notifiers[name] = {'cfg': cfg, 'tpl': None}

                if 'template' in cfg['general']:
                    with open(cfg['general']['template'], 'r') as t:
                        self.notifiers[name]['tpl'] = t.read()

                uri = urisup.uri_help_split(cfg['general']['uri'])
                if uri[0] in ('http', 'https'):
                    self.notifiers[name]['func'] = self._push_notify_http
                elif uri[0] == 'redis':
                    self.notifiers[name]['func'] = self._push_notify_redis
                else:
                    raise NotImplementedError("unsupported notifier: %r" % uri[0])

            if f:
                f.close()

    def _push_notifies(self, xvars):
        nvars                 = xvars.copy()
        nvars['_VARS_']       = xvars.copy()
        nvars['_SERVER_ID_']  = self.server_id
        nvars['_TIMESTAMP_']  = time.time()

        for name, notifier in self.notifiers.iteritems():
            if not notifier['cfg']['general'].get('enabled', True):
                continue

            if notifier['tpl']:
                tpl = json.loads(Template(notifier['tpl'],
                                          imports = ['import json',
                                                     'from escapejson import escapejson']).render(**nvars))
            else:
                tpl = None

            cfg                   = notifier['cfg'].copy()
            cfg['general']['uri'] = Template(cfg['general']['uri']).render(**nvars)

            notifier['func'](name, cfg, tpl)

    def _push_notify_http(self, name, cfg, tpl):
        (headers, payload) = ({}, {})

        if tpl:
            if 'headers' in tpl:
                headers = tpl['headers']

            if 'payload' in tpl:
                payload = tpl['payload']

        try:
            r = requests.post(cfg['general']['uri'],
                              headers = headers,
                              data    = payload)

            if 200 <= r.status_code < 300:
                LOG.info("notification pushed: %r", name)
                return True

            LOG.error("unable to push notification: %r: %r", name, r.text)
        except Exception, e:
            LOG.error("unable to push notification %r: %r", name, e)

    def _push_notify_redis(self, name, cfg, tpl):
        config        = {'general':
                          {'redis':
                            {'notifier': cfg['general'].get('options') or {}}}}
        config['general']['redis']['notifier']['url'] = cfg['general']['uri']

        if not tpl:
            LOG.error("missing redis template for %r", name)
            return

        try:
            adapter_redis = DWhoAdapterRedis(config, prefix = 'notifier')
            adapter_redis.set_key(tpl['key'], json.dumps(tpl['value']))
        except Exception, e:
            LOG.error("unable to push notification %r: %r", name, e)
        else:
            LOG.info("notification pushed: %r", name)


    NOTIFY_SCHEMA = xys.load("""
    uid!-: !!str
    backup!-: !!str
    type!-: !!str
    full*-: !!str
    partial*: !!bool False
    status!-: !~~enum(processing,failed,successful)
    backup_started*-: !!str
    backup_ended*-: !!str
    prepare_started*-: !!str
    prepare_ended*-: !!str
    compressed*: !!bool False
    compressed_with*-: !!str
    binary_log_name*-: !!str
    master_binary_log_name*-: !!str
    hostname!-: !!str
    errors*: [ !!str ]
    """)

    def notify(self, request):
        args = request.payload_params()

        if not xys.validate(args, self.NOTIFY_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            if args['type'] not in ('full', 'incr'):
                return

            xvars = args.copy()
            xvars['total_seconds'] = None

            if args.get('backup_started') and args.get('backup_ended'):
                if args['backup_started'].count('_') == 4:
                    xformat = '%Y_%m_%d-%H_%M_%S'
                else:
                    xformat = '%Y-%m-%d_%H-%M-%S'

                started                = datetime.strptime(args['backup_started'], xformat)
                ended                  = datetime.strptime(args['backup_ended'], xformat)
                xvars['total_seconds'] = (ended - started).total_seconds()

            DWHO_SHARED.set('xtrargon', 'state', args)

            self._push_notifies(xvars)

            return xvars
        except HttpReqErrJson:
            raise
        except Exception, e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(XtrargonBackupModule())
    _start()
